from quilt3.data_transfer import S3Api, S3ClientProvider, get_bytes, list_url
from quilt3.util import PhysicalKey

from .base import PackageRegistryV1, PackageRegistryV2


def s3_list_objects(**kwargs):
    s3_client = S3ClientProvider().find_correct_client(S3Api.LIST_OBJECTS_V2, kwargs['Bucket'], kwargs)
    return s3_client.get_paginator('list_objects_v2').paginate(**kwargs)


def delete_url_recursively(src: PhysicalKey):
    s3_client = S3ClientProvider().standard_client
    for resp in s3_list_objects(Bucket=src.bucket, Prefix=src.path):
        for key in resp.get('Contents', ()):
            s3_client.delete_object(Bucket=src.bucket, Key=key['Key'])


class S3PackageRegistryV1(PackageRegistryV1):
    def list_packages(self):
        prev_pkg = None
        for path, _ in list_url(self.pointers_global_dir):
            pkg = path.rpartition('/')[0]
            # A package can have multiple versions, but we should only return the name once.
            if pkg != prev_pkg:
                prev_pkg = pkg
                yield pkg

    def list_package_pointers(self, pkg_name: str):
        package_dir = self.pointers_dir(pkg_name)
        for path, _ in list_url(package_dir):
            pkg_hash = get_bytes(package_dir.join(path))
            yield path, pkg_hash.decode().strip()

    def delete_package(self, pkg_name: str):
        delete_url_recursively(self.pointers_dir(pkg_name))


class S3PackageRegistryV2(PackageRegistryV2):
    def list_packages(self):
        prefix = self.manifests_global_dir.path
        prefix_len = len(prefix)
        for resp in s3_list_objects(
            Bucket=self.manifests_global_dir.bucket,
            Prefix=prefix,
            Delimiter='/',
        ):
            for obj in resp.get('CommonPrefixes', ()):
                yield obj['Prefix'][prefix_len:-1].replace('@', '/')

    list_package_pointers = S3PackageRegistryV1.list_package_pointers

    def list_package_versions_with_timestamps(self, pkg_name: str):
        manifest_dir_pk = self.manifests_package_dir(pkg_name)
        prefix = manifest_dir_pk.path
        s = slice(len(prefix), None)
        for response in s3_list_objects(Bucket=manifest_dir_pk.bucket, Prefix=prefix):
            for obj in response.get('Contents', ()):
                yield obj['LastModified'], self._top_hash_from_path(obj['Key'][s])

    def delete_package(self, pkg_name: str):
        delete_url_recursively(self.manifests_package_dir(pkg_name))
        delete_url_recursively(self.pointers_dir(pkg_name))


def get_package_registry(version: int):
    if version == 1:
        return S3PackageRegistryV1
    if version == 2:
        return S3PackageRegistryV2
    raise ValueError()



import re
import unittest
from unittest import mock

import jsonschema
import pytest
from tests.utils import QuiltTestCase

from quilt3 import Package, workflows


class WorkflowConfigConfigDataVersionSupportTest(unittest.TestCase):
    version = (1, 1, 1)
    supported_versions = [
        (1, 0, 0),
        (1, 0, 1),
        (1, 1, 0),
        (1, 1, 1),
    ]
    not_supported_versions = [
        (2, 0, 0),
        (2, 0, 1),
        (2, 1, 0),
        (2, 1, 1),
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.version = workflows.ConfigDataVersion._make(cls.version)
        cls.supported_versions = list(map(workflows.ConfigDataVersion._make, cls.supported_versions))
        cls.not_supported_versions = list(map(workflows.ConfigDataVersion._make, cls.not_supported_versions))

    def setUp(self):
        super().setUp()

        patcher = mock.patch.object(workflows.WorkflowConfig, 'CONFIG_DATA_VERSION', self.version)
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_supported(self):
        assert self.supported_versions

        for v in self.supported_versions:
            with self.subTest(version=v):
                assert workflows.WorkflowConfig.is_supported_config_data_version(v) is True

    def test_not_supported(self):
        assert self.not_supported_versions

        for v in self.not_supported_versions:
            with self.subTest(version=v):
                assert workflows.WorkflowConfig.is_supported_config_data_version(v) is False


class WorkflowConfigGetVersionStrTest(unittest.TestCase):
    def test_version_str(self):
        data = {
            'version': '41.0.0',
        }
        assert workflows.WorkflowConfig.get_config_data_version_str(data) == '41.0.0'

    def test_version_object(self):
        data = {
            'version': {
                'ext1': '40.0.0',
                'base': '41.0.0',
                'ext2': '42.0.0',
            }
        }
        assert workflows.WorkflowConfig.get_config_data_version_str(data) == '41.0.0'


class ConfigDataVersionParseTest(unittest.TestCase):
    versions = {
        '1': (1, 0, 0),
        '1.0': (1, 0, 0),
        '1.0.0': (1, 0, 0),
        '1.0.1': (1, 0, 1),
        '1.1': (1, 1, 0),
        '1.1.1': (1, 1, 1),
    }

    def test(self):
        for version_string, expected_version in self.versions.items():
            expected_version = workflows.ConfigDataVersion._make(expected_version)
            with self.subTest(version_string=version_string, expected_version=expected_version):
                assert workflows.ConfigDataVersion.parse(version_string) == expected_version


class WorkflowValidatorTestMixin:
    def get_workflow_validator(self, **kwargs):
        return workflows.WorkflowValidator(
            **{
                'data_to_store': None,
                'is_message_required': False,
                'pkg_name_pattern': None,
                'metadata_validator': None,
                'entries_validator': None,
                **kwargs,
            }
        )


class WorkflowValidatorTest(unittest.TestCase, WorkflowValidatorTestMixin):
    JSON_SCHEMA_VALIDATOR_CLS = jsonschema.Draft7Validator

    def test_validate(self):
        pkg_name = 'test/name'
        msg = 'test message'
        meta = {'some': 'meta'}
        pkg = Package()
        pkg.set_meta(meta)

        workflow_validator = self.get_workflow_validator(data_to_store=mock.sentinel.data_to_store)
        methods_to_mock = (
            'validate_name',
            'validate_message',
            'validate_metadata',
            'validate_entries',
        )
        with mock.patch.multiple(workflows.WorkflowValidator, **dict.fromkeys(methods_to_mock, mock.DEFAULT)) as mocks:
            assert workflow_validator.validate(
                name=pkg_name,
                pkg=pkg,
                message=msg,
            ) is mock.sentinel.data_to_store

            mocks['validate_name'].assert_called_once_with(pkg_name)
            mocks['validate_message'].assert_called_once_with(msg)
            mocks['validate_metadata'].assert_called_once_with(meta)
            mocks['validate_entries'].assert_called_once_with(pkg)

    def test_validate_name_noop(self):
        workflow_validator = self.get_workflow_validator(pkg_name_pattern=None)
        workflow_validator.validate_name('foobar')

    def test_validate_name(self):
        workflow_validator = self.get_workflow_validator(pkg_name_pattern=re.compile(r'oob'))
        workflow_validator.validate_name('foobar')

    def test_validate_name_fail(self):
        workflow_validator = self.get_workflow_validator(pkg_name_pattern=re.compile(r'^oob'))
        with pytest.raises(workflows.WorkflowValidationError):
            workflow_validator.validate_name('foobar')

    def test_validate_message_not_required(self):
        workflow_validator = self.get_workflow_validator(is_message_required=False)
        for msg in (
            None,
            '',
            'message',
        ):
            with self.subTest(message=msg):
                workflow_validator.validate_message(msg)

    def test_validate_message_required(self):
        workflow_validator = self.get_workflow_validator(is_message_required=True)
        workflow_validator.validate_message('message')

    def test_validate_message_required_fail(self):
        workflow_validator = self.get_workflow_validator(is_message_required=True)
        for msg in (
            None,
            '',
        ):
            with self.subTest(message=msg):
                with pytest.raises(workflows.WorkflowValidationError):
                    workflow_validator.validate_message(msg)

    def test_validate_metadata_noop(self):
        workflow_validator = self.get_workflow_validator()
        workflow_validator.validate_metadata({})

    def test_validate_metadata(self):
        workflow_validator = self.get_workflow_validator(metadata_validator=self.JSON_SCHEMA_VALIDATOR_CLS(True))
        workflow_validator.validate_metadata({})

    def test_validate_metadata_fail(self):
        workflow_validator = self.get_workflow_validator(metadata_validator=self.JSON_SCHEMA_VALIDATOR_CLS(False))
        with pytest.raises(workflows.WorkflowValidationError):
            workflow_validator.validate_metadata({})

    @mock.patch.object(workflows.WorkflowValidator, 'get_pkg_entries_for_validation')
    def test_validate_pkg_entries_noop(self, get_pkg_entries_for_validation_mock):
        workflow_validator = self.get_workflow_validator()
        workflow_validator.validate_entries(Package())

        get_pkg_entries_for_validation_mock.assert_not_called()

    @mock.patch.object(workflows.WorkflowValidator, 'get_pkg_entries_for_validation')
    def test_validate_pkg_entries(self, get_pkg_entries_for_validation_mock):
        pkg = Package()

        workflow_validator = self.get_workflow_validator(entries_validator=self.JSON_SCHEMA_VALIDATOR_CLS(True))
        workflow_validator.validate_entries(pkg)

        get_pkg_entries_for_validation_mock.assert_called_once_with(pkg)

    @mock.patch.object(workflows.WorkflowValidator, 'get_pkg_entries_for_validation')
    def test_validate_pkg_entries_fail(self, get_pkg_entries_for_validation_mock):
        pkg = Package()

        workflow_validator = self.get_workflow_validator(entries_validator=self.JSON_SCHEMA_VALIDATOR_CLS(False))
        with pytest.raises(workflows.WorkflowValidationError):
            workflow_validator.validate_entries(pkg)

        get_pkg_entries_for_validation_mock.assert_called_once_with(pkg)


class GetPkgEntriesForValidationTest(QuiltTestCase, WorkflowValidatorTestMixin):
    def test(self):
        entries_data = {
            'b/a': (bytes(1), None),
            'a/b': (bytes(2), {}),
            'c': (bytes(3), {"test": "test"}),
        }
        pkg = Package()
        for lk, (data, meta) in entries_data.items():
            pkg.set(lk, data, meta=meta)

        workflow_validator = self.get_workflow_validator()
        assert workflow_validator.get_pkg_entries_for_validation(pkg) == [
            {
                'logical_key': 'a/b',
                'size': 2,
                "meta": {},
            },
            {
                'logical_key': 'b/a',
                'size': 1,
                "meta": {},
            },
            {
                'logical_key': 'c',
                'size': 3,
                "meta": {"test": "test"},
            },
        ]
